@@ -6,22 +6,26 @@ information to the adapters if needed.
 """
 
 # Copyright (C) 2020 The Psycopg Team
-from enum import Enum
-from typing import Any, Dict, Iterator, Optional, overload
-from typing import Sequence, Tuple, Type, TypeVar, Union, TYPE_CHECKING
-from typing_extensions import TypeAlias
 
+from __future__ import annotations
+
+from typing import Any, Iterator, overload, Sequence, TYPE_CHECKING
+
+from . import sql
 from . import errors as e
-from .abc import AdaptContext
+from .abc import AdaptContext, Query
 from .rows import dict_row
+from ._compat import TypeAlias, TypeVar
+from ._typemod import TypeModifier
+from ._encodings import conn_encoding
 
 if TYPE_CHECKING:
     from .connection import Connection
     from .connection_async import AsyncConnection
-    from .sql import Identifier
+    from ._connection_base import BaseConnection
 
 T = TypeVar("T", bound="TypeInfo")
-RegistryKey: TypeAlias = Union[str, int, Tuple[type, int]]
+RegistryKey: TypeAlias = "str | int | tuple[type, int]"
 
 
 class TypeInfo:
@@ -39,12 +43,14 @@ class TypeInfo:
         *,
         regtype: str = "",
         delimiter: str = ",",
+        typemod: type[TypeModifier] = TypeModifier,
     ):
         self.name = name
         self.oid = oid
         self.array_oid = array_oid
         self.regtype = regtype or name
         self.delimiter = delimiter
+        self.typemod = typemod(oid)
 
     def __repr__(self) -> str:
         return (
@@ -55,43 +61,48 @@ class TypeInfo:
     @overload
     @classmethod
     def fetch(
-        cls: Type[T], conn: "Connection[Any]", name: Union[str, "Identifier"]
-    ) -> Optional[T]:
-        ...
+        cls: type[T], conn: Connection[Any], name: str | sql.Identifier
+    ) -> T | None: ...
 
     @overload
     @classmethod
     async def fetch(
-        cls: Type[T],
-        conn: "AsyncConnection[Any]",
-        name: Union[str, "Identifier"],
-    ) -> Optional[T]:
-        ...
+        cls: type[T], conn: AsyncConnection[Any], name: str | sql.Identifier
+    ) -> T | None: ...
 
     @classmethod
     def fetch(
-        cls: Type[T],
-        conn: "Union[Connection[Any], AsyncConnection[Any]]",
-        name: Union[str, "Identifier"],
+        cls: type[T], conn: BaseConnection[Any], name: str | sql.Identifier
     ) -> Any:
         """Query a system catalog to read information about a type."""
-        from .sql import Composable
+        from .connection import Connection
         from .connection_async import AsyncConnection
 
-        if isinstance(name, Composable):
+        if isinstance(name, sql.Composable):
             name = name.as_string(conn)
 
-        if isinstance(conn, AsyncConnection):
+        if isinstance(conn, Connection):
+            return cls._fetch(conn, name)
+        elif isinstance(conn, AsyncConnection):
             return cls._fetch_async(conn, name)
+        else:
+            raise TypeError(
+                f"expected Connection or AsyncConnection, got {type(conn).__name__}"
+            )
 
+    @classmethod
+    def _fetch(cls: type[T], conn: Connection[Any], name: str) -> T | None:
         # This might result in a nested transaction. What we want is to leave
         # the function with the connection in the state we found (either idle
         # or intrans)
         try:
-            with conn.transaction():
-                with conn.cursor(binary=True, row_factory=dict_row) as cur:
-                    cur.execute(cls._get_info_query(conn), {"name": name})
-                    recs = cur.fetchall()
+            from psycopg import Cursor
+
+            with conn.transaction(), Cursor(conn, row_factory=dict_row) as cur:
+                if conn_encoding(conn) == "ascii":
+                    cur.execute("set local client_encoding to utf8")
+                cur.execute(cls._get_info_query(conn), {"name": name})
+                recs = cur.fetchall()
         except e.UndefinedObject:
             return None
 
@@ -99,16 +110,15 @@ class TypeInfo:
 
     @classmethod
     async def _fetch_async(
-        cls: Type[T], conn: "AsyncConnection[Any]", name: str
-    ) -> Optional[T]:
-        """
-        Query a system catalog to read information about a type.
-
-        Similar to `fetch()` but can use an asynchronous connection.
-        """
+        cls: type[T], conn: AsyncConnection[Any], name: str
+    ) -> T | None:
         try:
+            from psycopg import AsyncCursor
+
             async with conn.transaction():
-                async with conn.cursor(binary=True, row_factory=dict_row) as cur:
+                async with AsyncCursor(conn, row_factory=dict_row) as cur:
+                    if conn_encoding(conn) == "ascii":
+                        await cur.execute("set local client_encoding to utf8")
                     await cur.execute(cls._get_info_query(conn), {"name": name})
                     recs = await cur.fetchall()
         except e.UndefinedObject:
@@ -118,8 +128,8 @@ class TypeInfo:
 
     @classmethod
     def _from_records(
-        cls: Type[T], name: str, recs: Sequence[Dict[str, Any]]
-    ) -> Optional[T]:
+        cls: type[T], name: str, recs: Sequence[dict[str, Any]]
+    ) -> T | None:
         if len(recs) == 1:
             return cls(**recs[0])
         elif not recs:
@@ -127,7 +137,7 @@ class TypeInfo:
         else:
             raise e.ProgrammingError(f"found {len(recs)} different types named {name}")
 
-    def register(self, context: Optional[AdaptContext] = None) -> None:
+    def register(self, context: AdaptContext | None = None) -> None:
         """
         Register the type information, globally or in the specified `!context`.
         """
@@ -146,188 +156,64 @@ class TypeInfo:
             register_array(self, context)
 
     @classmethod
-    def _get_info_query(
-        cls, conn: "Union[Connection[Any], AsyncConnection[Any]]"
-    ) -> str:
-        return """\
+    def _get_info_query(cls, conn: BaseConnection[Any]) -> Query:
+        return sql.SQL(
+            """\
 SELECT
     typname AS name, oid, typarray AS array_oid,
     oid::regtype::text AS regtype, typdelim AS delimiter
 FROM pg_type t
-WHERE t.oid = %(name)s::regtype
+WHERE t.oid = {regtype}
 ORDER BY t.oid
 """
+        ).format(regtype=cls._to_regtype(conn))
 
-    def _added(self, registry: "TypesRegistry") -> None:
+    @classmethod
+    def _has_to_regtype_function(cls, conn: BaseConnection[Any]) -> bool:
+        # to_regtype() introduced in PostgreSQL 9.4 and CockroachDB 22.2
+        info = conn.info
+        if info.vendor == "PostgreSQL":
+            return info.server_version >= 90400
+        elif info.vendor == "CockroachDB":
+            return info.server_version >= 220200
+        else:
+            return False
+
+    @classmethod
+    def _to_regtype(cls, conn: BaseConnection[Any]) -> sql.SQL:
+        # `to_regtype()` returns the type oid or NULL, unlike the :: operator,
+        # which returns the type or raises an exception, which requires
+        # a transaction rollback and leaves traces in the server logs.
+
+        if cls._has_to_regtype_function(conn):
+            return sql.SQL("to_regtype(%(name)s)")
+        else:
+            return sql.SQL("%(name)s::regtype")
+
+    def _added(self, registry: TypesRegistry) -> None:
         """Method called by the `!registry` when the object is added there."""
         pass
 
+    def get_type_display(self, oid: int | None = None, fmod: int | None = None) -> str:
+        parts = []
+        parts.append(self.name)
+        mod = self.typemod.get_modifier(fmod) if fmod is not None else ()
+        if mod:
+            parts.append(f"({','.join(map(str, mod))})")
 
-class RangeInfo(TypeInfo):
-    """Manage information about a range type."""
+        if oid == self.array_oid:
+            parts.append("[]")
 
-    __module__ = "psycopg.types.range"
+        return "".join(parts)
 
-    def __init__(
-        self,
-        name: str,
-        oid: int,
-        array_oid: int,
-        *,
-        regtype: str = "",
-        subtype_oid: int,
-    ):
-        super().__init__(name, oid, array_oid, regtype=regtype)
-        self.subtype_oid = subtype_oid
+    def get_display_size(self, fmod: int) -> int | None:
+        return self.typemod.get_display_size(fmod)
 
-    @classmethod
-    def _get_info_query(
-        cls, conn: "Union[Connection[Any], AsyncConnection[Any]]"
-    ) -> str:
-        return """\
-SELECT t.typname AS name, t.oid AS oid, t.typarray AS array_oid,
-    t.oid::regtype::text AS regtype,
-    r.rngsubtype AS subtype_oid
-FROM pg_type t
-JOIN pg_range r ON t.oid = r.rngtypid
-WHERE t.oid = %(name)s::regtype
-"""
+    def get_precision(self, fmod: int) -> int | None:
+        return self.typemod.get_precision(fmod)
 
-    def _added(self, registry: "TypesRegistry") -> None:
-        # Map ranges subtypes to info
-        registry._registry[RangeInfo, self.subtype_oid] = self
-
-
-class MultirangeInfo(TypeInfo):
-    """Manage information about a multirange type."""
-
-    # TODO: expose to multirange module once added
-    # __module__ = "psycopg.types.multirange"
-
-    def __init__(
-        self,
-        name: str,
-        oid: int,
-        array_oid: int,
-        *,
-        regtype: str = "",
-        range_oid: int,
-        subtype_oid: int,
-    ):
-        super().__init__(name, oid, array_oid, regtype=regtype)
-        self.range_oid = range_oid
-        self.subtype_oid = subtype_oid
-
-    @classmethod
-    def _get_info_query(
-        cls, conn: "Union[Connection[Any], AsyncConnection[Any]]"
-    ) -> str:
-        if conn.info.server_version < 140000:
-            raise e.NotSupportedError(
-                "multirange types are only available from PostgreSQL 14"
-            )
-        return """\
-SELECT t.typname AS name, t.oid AS oid, t.typarray AS array_oid,
-    t.oid::regtype::text AS regtype,
-    r.rngtypid AS range_oid, r.rngsubtype AS subtype_oid
-FROM pg_type t
-JOIN pg_range r ON t.oid = r.rngmultitypid
-WHERE t.oid = %(name)s::regtype
-"""
-
-    def _added(self, registry: "TypesRegistry") -> None:
-        # Map multiranges ranges and subtypes to info
-        registry._registry[MultirangeInfo, self.range_oid] = self
-        registry._registry[MultirangeInfo, self.subtype_oid] = self
-
-
-class CompositeInfo(TypeInfo):
-    """Manage information about a composite type."""
-
-    __module__ = "psycopg.types.composite"
-
-    def __init__(
-        self,
-        name: str,
-        oid: int,
-        array_oid: int,
-        *,
-        regtype: str = "",
-        field_names: Sequence[str],
-        field_types: Sequence[int],
-    ):
-        super().__init__(name, oid, array_oid, regtype=regtype)
-        self.field_names = field_names
-        self.field_types = field_types
-        # Will be set by register() if the `factory` is a type
-        self.python_type: Optional[type] = None
-
-    @classmethod
-    def _get_info_query(
-        cls, conn: "Union[Connection[Any], AsyncConnection[Any]]"
-    ) -> str:
-        return """\
-SELECT
-    t.typname AS name, t.oid AS oid, t.typarray AS array_oid,
-    t.oid::regtype::text AS regtype,
-    coalesce(a.fnames, '{}') AS field_names,
-    coalesce(a.ftypes, '{}') AS field_types
-FROM pg_type t
-LEFT JOIN (
-    SELECT
-        attrelid,
-        array_agg(attname) AS fnames,
-        array_agg(atttypid) AS ftypes
-    FROM (
-        SELECT a.attrelid, a.attname, a.atttypid
-        FROM pg_attribute a
-        JOIN pg_type t ON t.typrelid = a.attrelid
-        WHERE t.oid = %(name)s::regtype
-        AND a.attnum > 0
-        AND NOT a.attisdropped
-        ORDER BY a.attnum
-    ) x
-    GROUP BY attrelid
-) a ON a.attrelid = t.typrelid
-WHERE t.oid = %(name)s::regtype
-"""
-
-
-class EnumInfo(TypeInfo):
-    """Manage information about an enum type."""
-
-    __module__ = "psycopg.types.enum"
-
-    def __init__(
-        self,
-        name: str,
-        oid: int,
-        array_oid: int,
-        labels: Sequence[str],
-    ):
-        super().__init__(name, oid, array_oid)
-        self.labels = labels
-        # Will be set by register_enum()
-        self.enum: Optional[Type[Enum]] = None
-
-    @classmethod
-    def _get_info_query(
-        cls, conn: "Union[Connection[Any], AsyncConnection[Any]]"
-    ) -> str:
-        return """\
-SELECT name, oid, array_oid, array_agg(label) AS labels
-FROM (
-    SELECT
-        t.typname AS name, t.oid AS oid, t.typarray AS array_oid,
-        e.enumlabel AS label
-    FROM pg_type t
-    LEFT JOIN  pg_enum e
-    ON e.enumtypid = t.oid
-    WHERE t.oid = %(name)s::regtype
-    ORDER BY e.enumsortorder
-) x
-GROUP BY name, oid, array_oid
-"""
+    def get_scale(self, fmod: int) -> int | None:
+        return self.typemod.get_scale(fmod)
 
 
 class TypesRegistry:
@@ -337,8 +223,8 @@ class TypesRegistry:
 
     __module__ = "psycopg.types"
 
-    def __init__(self, template: Optional["TypesRegistry"] = None):
-        self._registry: Dict[RegistryKey, TypeInfo]
+    def __init__(self, template: TypesRegistry | None = None):
+        self._registry: dict[RegistryKey, TypeInfo]
 
         # Make a shallow copy: it will become a proper copy if the registry
         # is edited.
@@ -375,12 +261,10 @@ class TypesRegistry:
                 yield t
 
     @overload
-    def __getitem__(self, key: Union[str, int]) -> TypeInfo:
-        ...
+    def __getitem__(self, key: str | int) -> TypeInfo: ...
 
     @overload
-    def __getitem__(self, key: Tuple[Type[T], int]) -> T:
-        ...
+    def __getitem__(self, key: tuple[type[T], int]) -> T: ...
 
     def __getitem__(self, key: RegistryKey) -> TypeInfo:
         """
@@ -401,14 +285,12 @@ class TypesRegistry:
             raise KeyError(f"couldn't find the type {key!r} in the types registry")
 
     @overload
-    def get(self, key: Union[str, int]) -> Optional[TypeInfo]:
-        ...
+    def get(self, key: str | int) -> TypeInfo | None: ...
 
     @overload
-    def get(self, key: Tuple[Type[T], int]) -> Optional[T]:
-        ...
+    def get(self, key: tuple[type[T], int]) -> T | None: ...
 
-    def get(self, key: RegistryKey) -> Optional[TypeInfo]:
+    def get(self, key: RegistryKey) -> TypeInfo | None:
         """
         Return info about a type, specified by name or oid
 
@@ -437,7 +319,7 @@ class TypesRegistry:
         else:
             return t.oid
 
-    def get_by_subtype(self, cls: Type[T], subtype: Union[int, str]) -> Optional[T]:
+    def get_by_subtype(self, cls: type[T], subtype: int | str) -> T | None:
         """
         Return info about a `TypeInfo` subclass by its element name or oid.
 
