@@ -18,16 +18,225 @@ DBAPI-defined Exceptions are defined in the following hierarchy::
 
 # Copyright (C) 2020 The Psycopg Team
 
-from typing import Any, Dict, Optional, Sequence, Tuple, Type, Union
-from typing_extensions import TypeAlias
+from __future__ import annotations
+
+from dataclasses import dataclass, field, fields
+from typing import Any, Callable, NoReturn, Sequence, TYPE_CHECKING
+from asyncio import CancelledError
 
 from .pq.abc import PGconn, PGresult
-from .pq._enums import DiagnosticField
-from ._compat import TypeGuard
+from .pq._enums import ConnStatus, DiagnosticField, PipelineStatus, TransactionStatus
+from ._compat import TypeAlias, TypeGuard
 
-ErrorInfo: TypeAlias = Union[None, PGresult, Dict[int, Optional[bytes]]]
+if TYPE_CHECKING:
+    from .pq.misc import PGnotify, ConninfoOption
 
-_sqlcodes: Dict[str, "Type[Error]"] = {}
+ErrorInfo: TypeAlias = "PGresult | dict[int, bytes | None] | None"
+
+_sqlcodes: dict[str, type[Error]] = {}
+
+
+@dataclass
+class FinishedPGconn:
+    """Finished libpq connection.
+
+    Attributes are set from a real `~psycopg.pq.PGconn` but any operations will
+    raise an `~psycopg.OperationalError`.
+    """
+
+    info: list[ConninfoOption] = field(default_factory=list)
+
+    db: bytes = b""
+    user: bytes = b""
+    password: bytes = b""
+    host: bytes = b""
+    hostaddr: bytes = b""
+    port: bytes = b""
+    tty: bytes = b""
+    options: bytes = b""
+    status: int = ConnStatus.BAD.value
+    transaction_status: int = TransactionStatus.UNKNOWN.value
+    pipeline_status: int = PipelineStatus.OFF.value
+
+    error_message: bytes = b""
+    _encoding: str = "utf-8"
+    server_version: int = 0
+
+    backend_pid: int = 0
+    needs_password: bool = False
+    used_password: bool = False
+    ssl_in_use: bool = False
+
+    nonblocking: int = 0
+
+    notice_handler: Callable[[PGresult], None] | None = None
+    notify_handler: Callable[[PGnotify], None] | None = None
+
+    @staticmethod
+    def _raise() -> NoReturn:
+        raise OperationalError("the connection is closed")
+
+    @classmethod
+    def connect(cls, *args: Any) -> NoReturn:
+        raise TypeError(f"{cls} is unusable")
+
+    @classmethod
+    def connect_start(cls, *args: Any) -> NoReturn:
+        raise TypeError(f"{cls} is unusable")
+
+    def connect_poll(self) -> NoReturn:
+        self._raise()
+
+    def finish(self) -> None:
+        pass
+
+    def reset(self) -> NoReturn:
+        self._raise()
+
+    def get_error_message(self, encoding: str = "") -> str:
+        return "the connection is closed"
+
+    def reset_start(self) -> NoReturn:
+        self._raise()
+
+    def reset_poll(self) -> NoReturn:
+        self._raise()
+
+    @classmethod
+    def ping(cls, *args: Any) -> NoReturn:
+        raise TypeError(f"{cls} is unusable")
+
+    def parameter_status(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    @property
+    def socket(self) -> NoReturn:
+        self._raise()
+
+    def exec_(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def send_query(self, *args: Any) -> None:
+        self._raise()
+
+    def exec_params(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def send_query_params(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def send_prepare(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def send_query_prepared(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def prepare(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def exec_prepared(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def describe_prepared(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def send_describe_prepared(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def describe_portal(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def send_describe_portal(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def close_prepared(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def send_close_prepared(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def close_portal(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def send_close_portal(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def get_result(self) -> NoReturn:
+        self._raise()
+
+    def consume_input(self) -> NoReturn:
+        self._raise()
+
+    def is_busy(self) -> NoReturn:
+        self._raise()
+
+    def flush(self) -> NoReturn:
+        self._raise()
+
+    def set_single_row_mode(self) -> NoReturn:
+        self._raise()
+
+    def set_chunked_rows_mode(self, size: int) -> NoReturn:
+        self._raise()
+
+    def cancel_conn(self) -> NoReturn:
+        self._raise()
+
+    def get_cancel(self) -> NoReturn:
+        self._raise()
+
+    def notifies(self) -> NoReturn:
+        self._raise()
+
+    def put_copy_data(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def put_copy_end(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def get_copy_data(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def trace(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def set_trace_flags(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def untrace(self) -> NoReturn:
+        self._raise()
+
+    def encrypt_password(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def change_password(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def make_empty_result(self, *args: Any) -> NoReturn:
+        self._raise()
+
+    def enter_pipeline_mode(self) -> NoReturn:
+        self._raise()
+
+    def exit_pipeline_mode(self) -> NoReturn:
+        self._raise()
+
+    def pipeline_sync(self) -> NoReturn:
+        self._raise()
+
+    def send_flush_request(self) -> NoReturn:
+        self._raise()
+
+
+def finish_pgconn(pgconn: PGconn) -> PGconn:
+    args = {}
+    for f in fields(FinishedPGconn):
+        try:
+            args[f.name] = getattr(pgconn, f.name)
+        except Exception:
+            pass
+    pgconn.finish()
+    return FinishedPGconn(**args)
 
 
 class Warning(Exception):
@@ -52,14 +261,14 @@ class Error(Exception):
 
     __module__ = "psycopg"
 
-    sqlstate: Optional[str] = None
+    sqlstate: str | None = None
 
     def __init__(
         self,
         *args: Sequence[Any],
         info: ErrorInfo = None,
         encoding: str = "utf-8",
-        pgconn: Optional[PGconn] = None
+        pgconn: PGconn | None = None,
     ):
         super().__init__(*args)
         self._info = info
@@ -71,29 +280,29 @@ class Error(Exception):
             self.sqlstate = self.diag.sqlstate
 
     @property
-    def pgconn(self) -> Optional[PGconn]:
+    def pgconn(self) -> PGconn | None:
         """The connection object, if the error was raised from a connection attempt.
 
-        :rtype: Optional[psycopg.pq.PGconn]
+        :rtype: psycopg.pq.PGconn | None
         """
         return self._pgconn if self._pgconn else None
 
     @property
-    def pgresult(self) -> Optional[PGresult]:
+    def pgresult(self) -> PGresult | None:
         """The result object, if the exception was raised after a failed query.
 
-        :rtype: Optional[psycopg.pq.PGresult]
+        :rtype: psycopg.pq.PGresult | None
         """
         return self._info if _is_pgresult(self._info) else None
 
     @property
-    def diag(self) -> "Diagnostic":
+    def diag(self) -> Diagnostic:
         """
         A `Diagnostic` object to inspect details of the errors from the database.
         """
         return Diagnostic(self._info, encoding=self._encoding)
 
-    def __reduce__(self) -> Union[str, Tuple[Any, ...]]:
+    def __reduce__(self) -> str | tuple[Any, ...]:
         res = super().__reduce__()
         if isinstance(res, tuple) and len(res) >= 3:
             # To make the exception picklable
@@ -118,7 +327,7 @@ class DatabaseError(Error):
 
     __module__ = "psycopg"
 
-    def __init_subclass__(cls, code: Optional[str] = None, name: Optional[str] = None):
+    def __init_subclass__(cls, code: str | None = None, name: str | None = None):
         if code:
             _sqlcodes[code] = cls
             cls.sqlstate = code
@@ -200,6 +409,15 @@ class ConnectionTimeout(OperationalError):
     """
 
 
+class CancellationTimeout(OperationalError):
+    """
+    Exception raised on timeout of connection's
+    `~psycopg.Connection.cancel_safe()` method.
+
+    Subclass of `~psycopg.OperationalError`.
+    """
+
+
 class PipelineAborted(OperationalError):
     """
     Raised when a operation fails because the current pipeline is in aborted state.
@@ -216,78 +434,78 @@ class Diagnostic:
         self._encoding = encoding
 
     @property
-    def severity(self) -> Optional[str]:
+    def severity(self) -> str | None:
         return self._error_message(DiagnosticField.SEVERITY)
 
     @property
-    def severity_nonlocalized(self) -> Optional[str]:
+    def severity_nonlocalized(self) -> str | None:
         return self._error_message(DiagnosticField.SEVERITY_NONLOCALIZED)
 
     @property
-    def sqlstate(self) -> Optional[str]:
+    def sqlstate(self) -> str | None:
         return self._error_message(DiagnosticField.SQLSTATE)
 
     @property
-    def message_primary(self) -> Optional[str]:
+    def message_primary(self) -> str | None:
         return self._error_message(DiagnosticField.MESSAGE_PRIMARY)
 
     @property
-    def message_detail(self) -> Optional[str]:
+    def message_detail(self) -> str | None:
         return self._error_message(DiagnosticField.MESSAGE_DETAIL)
 
     @property
-    def message_hint(self) -> Optional[str]:
+    def message_hint(self) -> str | None:
         return self._error_message(DiagnosticField.MESSAGE_HINT)
 
     @property
-    def statement_position(self) -> Optional[str]:
+    def statement_position(self) -> str | None:
         return self._error_message(DiagnosticField.STATEMENT_POSITION)
 
     @property
-    def internal_position(self) -> Optional[str]:
+    def internal_position(self) -> str | None:
         return self._error_message(DiagnosticField.INTERNAL_POSITION)
 
     @property
-    def internal_query(self) -> Optional[str]:
+    def internal_query(self) -> str | None:
         return self._error_message(DiagnosticField.INTERNAL_QUERY)
 
     @property
-    def context(self) -> Optional[str]:
+    def context(self) -> str | None:
         return self._error_message(DiagnosticField.CONTEXT)
 
     @property
-    def schema_name(self) -> Optional[str]:
+    def schema_name(self) -> str | None:
         return self._error_message(DiagnosticField.SCHEMA_NAME)
 
     @property
-    def table_name(self) -> Optional[str]:
+    def table_name(self) -> str | None:
         return self._error_message(DiagnosticField.TABLE_NAME)
 
     @property
-    def column_name(self) -> Optional[str]:
+    def column_name(self) -> str | None:
         return self._error_message(DiagnosticField.COLUMN_NAME)
 
     @property
-    def datatype_name(self) -> Optional[str]:
+    def datatype_name(self) -> str | None:
         return self._error_message(DiagnosticField.DATATYPE_NAME)
 
     @property
-    def constraint_name(self) -> Optional[str]:
+    def constraint_name(self) -> str | None:
         return self._error_message(DiagnosticField.CONSTRAINT_NAME)
 
     @property
-    def source_file(self) -> Optional[str]:
+    def source_file(self) -> str | None:
         return self._error_message(DiagnosticField.SOURCE_FILE)
 
     @property
-    def source_line(self) -> Optional[str]:
+    def source_line(self) -> str | None:
         return self._error_message(DiagnosticField.SOURCE_LINE)
 
     @property
-    def source_function(self) -> Optional[str]:
+    def source_function(self) -> str | None:
         return self._error_message(DiagnosticField.SOURCE_FUNCTION)
 
-    def _error_message(self, field: DiagnosticField) -> Optional[str]:
+    def _error_message(self, field: DiagnosticField) -> str | None:
         if self._info:
             if isinstance(self._info, dict):
                 val = self._info.get(field)
@@ -299,7 +517,7 @@ class Diagnostic:
 
         return None
 
-    def __reduce__(self) -> Union[str, Tuple[Any, ...]]:
+    def __reduce__(self) -> str | tuple[Any, ...]:
         res = super().__reduce__()
         if isinstance(res, tuple) and len(res) >= 3:
             res[2]["_info"] = _info_to_dict(self._info)
@@ -318,7 +536,7 @@ def _info_to_dict(info: ErrorInfo) -> ErrorInfo:
         return info
 
 
-def lookup(sqlstate: str) -> Type[Error]:
+def lookup(sqlstate: str) -> type[Error]:
     """Lookup an error code or `constant name`__ and return its exception class.
 
     Raise `!KeyError` if the code is not found.
@@ -330,15 +548,9 @@ def lookup(sqlstate: str) -> Type[Error]:
 
 
 def error_from_result(result: PGresult, encoding: str = "utf-8") -> Error:
-    from psycopg import pq
-
     state = result.error_field(DiagnosticField.SQLSTATE) or b""
-    cls = _class_for_state(state.decode("ascii"))
-    return cls(
-        pq.error_message(result, encoding=encoding),
-        info=result,
-        encoding=encoding,
-    )
+    cls = _class_for_state(state.decode("utf-8", "replace"))
+    return cls(result.get_error_message(encoding), info=result, encoding=encoding)
 
 
 def _is_pgresult(info: ErrorInfo) -> TypeGuard[PGresult]:
@@ -347,14 +559,14 @@ def _is_pgresult(info: ErrorInfo) -> TypeGuard[PGresult]:
     return hasattr(info, "error_field")
 
 
-def _class_for_state(sqlstate: str) -> Type[Error]:
+def _class_for_state(sqlstate: str) -> type[Error]:
     try:
         return lookup(sqlstate)
     except KeyError:
         return get_base_exception(sqlstate)
 
 
-def get_base_exception(sqlstate: str) -> Type[Error]:
+def get_base_exception(sqlstate: str) -> type[Error]:
     return (
         _base_exc_map.get(sqlstate[:2])
         or _base_exc_map.get(sqlstate[:1])
@@ -891,6 +1103,10 @@ class IdleInTransactionSessionTimeout(InternalError,
     code='25P03', name='IDLE_IN_TRANSACTION_SESSION_TIMEOUT'):
     pass
 
+class TransactionTimeout(InternalError,
+    code='25P04', name='TRANSACTION_TIMEOUT'):
+    pass
+
 
 # Class 26 - Invalid SQL Statement Name
 
@@ -1533,3 +1749,11 @@ class IndexCorrupted(InternalError,
 
 # autogenerated: end
 # fmt: on
+
+# Don't show a complete traceback upon raising these exception.
+# Usually the traceback starts from internal functions (for instance in the
+# server communication callbacks) but, for the end user, it's more important
+# to get the high level information about where the exception was raised, for
+# instance in a certain `Cursor.execute()`.
+
+_NO_TRACEBACK = (Error, KeyboardInterrupt, CancelledError)
