@@ -8,13 +8,13 @@ These functions are designed to consume the generators returned by the
 
 # Copyright (C) 2020 The Psycopg Team
 
+from __future__ import annotations
 
 import os
 import sys
 import select
 import logging
 import selectors
-from typing import Optional
 from asyncio import get_event_loop, wait_for, Event, TimeoutError
 from selectors import DefaultSelector
 
@@ -26,6 +26,7 @@ from ._cmodule import _psycopg
 WAIT_R = Wait.R
 WAIT_W = Wait.W
 WAIT_RW = Wait.RW
+READY_NONE = Ready.NONE
 READY_R = Ready.R
 READY_W = Ready.W
 READY_RW = Ready.RW
@@ -33,16 +34,15 @@ READY_RW = Ready.RW
 logger = logging.getLogger(__name__)
 
 
-def wait_selector(gen: PQGen[RV], fileno: int, timeout: Optional[float] = None) -> RV:
+def wait_selector(gen: PQGen[RV], fileno: int, interval: float | None = None) -> RV:
     """
     Wait for a generator using the best strategy available.
 
     :param gen: a generator performing database operations and yielding
         `Ready` values when it would block.
     :param fileno: the file descriptor to wait on.
-    :param timeout: timeout (in seconds) to check for other interrupt, e.g.
-        to allow Ctrl-C.
-    :type timeout: float
+    :param interval: interval (in seconds) to check for other interrupt, e.g.
+        to allow Ctrl-C. If zero or None, wait indefinitely.
     :return: whatever `!gen` returns on completion.
 
     Consume `!gen`, scheduling `fileno` for completion when it is reported to
@@ -51,31 +51,31 @@ def wait_selector(gen: PQGen[RV], fileno: int, timeout: Optional[float] = None)
     try:
         s = next(gen)
         with DefaultSelector() as sel:
+            sel.register(fileno, s)
             while True:
-                sel.register(fileno, s)
-                rlist = None
-                while not rlist:
-                    rlist = sel.select(timeout=timeout)
+                rlist = sel.select(timeout=interval)
+                if not rlist:
+                    gen.send(READY_NONE)
+                    continue
+
                 sel.unregister(fileno)
-                # note: this line should require a cast, but mypy doesn't complain
-                ready: Ready = rlist[0][1]
-                assert s & ready
+                ready = rlist[0][1]
                 s = gen.send(ready)
+                sel.register(fileno, s)
 
     except StopIteration as ex:
-        rv: RV = ex.args[0] if ex.args else None
+        rv: RV = ex.value
         return rv
 
 
-def wait_conn(gen: PQGenConn[RV], timeout: Optional[float] = None) -> RV:
+def wait_conn(gen: PQGenConn[RV], interval: float | None = None) -> RV:
     """
     Wait for a connection generator using the best strategy available.
 
     :param gen: a generator performing database operations and yielding
         (fd, `Ready`) pairs when it would block.
-    :param timeout: timeout (in seconds) to check for other interrupt, e.g.
+    :param interval: interval (in seconds) to check for other interrupt, e.g.
         to allow Ctrl-C. If zero or None, wait indefinitely.
-    :type timeout: float
     :return: whatever `!gen` returns on completion.
 
     Behave like in `wait()`, but take the fileno to wait from the generator
@@ -83,32 +83,35 @@ def wait_conn(gen: PQGenConn[RV], timeout: Optional[float] = None) -> RV:
     """
     try:
         fileno, s = next(gen)
-        if not timeout:
-            timeout = None
+        if not interval:
+            interval = None
         with DefaultSelector() as sel:
+            sel.register(fileno, s)
             while True:
-                sel.register(fileno, s)
-                rlist = sel.select(timeout=timeout)
-                sel.unregister(fileno)
+                rlist = sel.select(timeout=interval)
                 if not rlist:
-                    raise e.ConnectionTimeout("connection timeout expired")
-                ready: Ready = rlist[0][1]  # type: ignore[assignment]
+                    gen.send(READY_NONE)
+                    continue
+
+                sel.unregister(fileno)
+                ready = rlist[0][1]
                 fileno, s = gen.send(ready)
+                sel.register(fileno, s)
 
     except StopIteration as ex:
-        rv: RV = ex.args[0] if ex.args else None
+        rv: RV = ex.value
         return rv
 
 
-async def wait_async(
-    gen: PQGen[RV], fileno: int, timeout: Optional[float] = None
-) -> RV:
+async def wait_async(gen: PQGen[RV], fileno: int, interval: float | None = None) -> RV:
     """
     Coroutine waiting for a generator to complete.
 
     :param gen: a generator performing database operations and yielding
         `Ready` values when it would block.
     :param fileno: the file descriptor to wait on.
+    :param interval: interval (in seconds) to check for other interrupt, e.g.
+        to allow Ctrl-C. If None, wait indefinitely.
     :return: whatever `!gen` returns on completion.
 
     Behave like in `wait()`, but exposing an `asyncio` interface.
@@ -117,12 +120,12 @@ async def wait_async(
     # Not sure this is the best implementation but it's a start.
     ev = Event()
     loop = get_event_loop()
-    ready: Ready
+    ready: int
     s: Wait
 
     def wakeup(state: Ready) -> None:
         nonlocal ready
-        ready |= state  # type: ignore[assignment]
+        ready |= state
         ev.set()
 
     try:
@@ -133,19 +136,19 @@ async def wait_async(
             if not reader and not writer:
                 raise e.InternalError(f"bad poll status: {s}")
             ev.clear()
-            ready = 0  # type: ignore[assignment]
+            ready = 0
             if reader:
                 loop.add_reader(fileno, wakeup, READY_R)
             if writer:
                 loop.add_writer(fileno, wakeup, READY_W)
             try:
-                if timeout is None:
-                    await ev.wait()
-                else:
+                if interval is not None:
                     try:
-                        await wait_for(ev.wait(), timeout)
+                        await wait_for(ev.wait(), interval)
                     except TimeoutError:
                         pass
+                else:
+                    await ev.wait()
             finally:
                 if reader:
                     loop.remove_reader(fileno)
@@ -153,18 +156,21 @@ async def wait_async(
                     loop.remove_writer(fileno)
             s = gen.send(ready)
 
+    except OSError as ex:
+        # Assume the connection was closed
+        raise e.OperationalError(str(ex))
     except StopIteration as ex:
-        rv: RV = ex.args[0] if ex.args else None
+        rv: RV = ex.value
         return rv
 
 
-async def wait_conn_async(gen: PQGenConn[RV], timeout: Optional[float] = None) -> RV:
+async def wait_conn_async(gen: PQGenConn[RV], interval: float | None = None) -> RV:
     """
     Coroutine waiting for a connection generator to complete.
 
     :param gen: a generator performing database operations and yielding
         (fd, `Ready`) pairs when it would block.
-    :param timeout: timeout (in seconds) to check for other interrupt, e.g.
+    :param interval: interval (in seconds) to check for other interrupt, e.g.
         to allow Ctrl-C. If zero or None, wait indefinitely.
     :return: whatever `!gen` returns on completion.
 
@@ -185,8 +191,6 @@ async def wait_conn_async(gen: PQGenConn[RV], timeout: Optional[float] = None) -
 
     try:
         fileno, s = next(gen)
-        if not timeout:
-            timeout = None
         while True:
             reader = s & WAIT_R
             writer = s & WAIT_W
@@ -199,7 +203,13 @@ async def wait_conn_async(gen: PQGenConn[RV], timeout: Optional[float] = None) -
             if writer:
                 loop.add_writer(fileno, wakeup, READY_W)
             try:
-                await wait_for(ev.wait(), timeout)
+                if interval:
+                    try:
+                        await wait_for(ev.wait(), interval)
+                    except TimeoutError:
+                        pass
+                else:
+                    await ev.wait()
             finally:
                 if reader:
                     loop.remove_reader(fileno)
@@ -207,18 +217,15 @@ async def wait_conn_async(gen: PQGenConn[RV], timeout: Optional[float] = None) -
                     loop.remove_writer(fileno)
             fileno, s = gen.send(ready)
 
-    except TimeoutError:
-        raise e.ConnectionTimeout("connection timeout expired")
-
     except StopIteration as ex:
-        rv: RV = ex.args[0] if ex.args else None
+        rv: RV = ex.value
         return rv
 
 
 # Specialised implementation of wait functions.
 
 
-def wait_select(gen: PQGen[RV], fileno: int, timeout: Optional[float] = None) -> RV:
+def wait_select(gen: PQGen[RV], fileno: int, interval: float | None = None) -> RV:
     """
     Wait for a generator using select where supported.
 
@@ -234,7 +241,7 @@ def wait_select(gen: PQGen[RV], fileno: int, timeout: Optional[float] = None) ->
                 fnlist if s & WAIT_R else empty,
                 fnlist if s & WAIT_W else empty,
                 fnlist,
-                timeout,
+                interval,
             )
             ready = 0
             if rl:
@@ -242,12 +249,13 @@ def wait_select(gen: PQGen[RV], fileno: int, timeout: Optional[float] = None) ->
             if wl:
                 ready |= READY_W
             if not ready:
+                gen.send(READY_NONE)
                 continue
-            # assert s & ready
-            s = gen.send(ready)  # type: ignore
+
+            s = gen.send(ready)
 
     except StopIteration as ex:
-        rv: RV = ex.args[0] if ex.args else None
+        rv: RV = ex.value
         return rv
 
 
@@ -262,7 +270,7 @@ else:
     _epoll_evmasks = {}
 
 
-def wait_epoll(gen: PQGen[RV], fileno: int, timeout: Optional[float] = None) -> RV:
+def wait_epoll(gen: PQGen[RV], fileno: int, interval: float | None = None) -> RV:
     """
     Wait for a generator using epoll where supported.
 
@@ -281,31 +289,29 @@ def wait_epoll(gen: PQGen[RV], fileno: int, timeout: Optional[float] = None) ->
     try:
         s = next(gen)
 
-        if timeout is None or timeout < 0:
-            timeout = 0
-        else:
-            timeout = int(timeout * 1000.0)
+        if interval is None or interval < 0:
+            interval = 0.0
 
         with select.epoll() as epoll:
             evmask = _epoll_evmasks[s]
             epoll.register(fileno, evmask)
             while True:
-                fileevs = None
-                while not fileevs:
-                    fileevs = epoll.poll(timeout)
+                fileevs = epoll.poll(interval)
+                if not fileevs:
+                    gen.send(READY_NONE)
+                    continue
                 ev = fileevs[0][1]
                 ready = 0
                 if ev & ~select.EPOLLOUT:
                     ready = READY_R
                 if ev & ~select.EPOLLIN:
                     ready |= READY_W
-                # assert s & ready
                 s = gen.send(ready)
                 evmask = _epoll_evmasks[s]
                 epoll.modify(fileno, evmask)
 
     except StopIteration as ex:
-        rv: RV = ex.args[0] if ex.args else None
+        rv: RV = ex.value
         return rv
 
 
@@ -319,7 +325,7 @@ else:
     _poll_evmasks = {}
 
 
-def wait_poll(gen: PQGen[RV], fileno: int, timeout: Optional[float] = None) -> RV:
+def wait_poll(gen: PQGen[RV], fileno: int, interval: float | None = None) -> RV:
     """
     Wait for a generator using poll where supported.
 
@@ -328,31 +334,32 @@ def wait_poll(gen: PQGen[RV], fileno: int, timeout: Optional[float] = None) -> R
     try:
         s = next(gen)
 
-        if timeout is None or timeout < 0:
-            timeout = 0
+        if interval is None or interval < 0:
+            interval = 0
         else:
-            timeout = int(timeout * 1000.0)
+            interval = int(interval * 1000.0)
 
         poll = select.poll()
         evmask = _poll_evmasks[s]
         poll.register(fileno, evmask)
         while True:
-            fileevs = None
-            while not fileevs:
-                fileevs = poll.poll(timeout)
+            fileevs = poll.poll(interval)
+            if not fileevs:
+                gen.send(READY_NONE)
+                continue
+
             ev = fileevs[0][1]
             ready = 0
             if ev & ~select.POLLOUT:
                 ready = READY_R
             if ev & ~select.POLLIN:
                 ready |= READY_W
-            # assert s & ready
             s = gen.send(ready)
             evmask = _poll_evmasks[s]
             poll.modify(fileno, evmask)
 
     except StopIteration as ex:
-        rv: RV = ex.args[0] if ex.args else None
+        rv: RV = ex.value
         return rv
 
 
