@@ -4,51 +4,44 @@ compatibility functions for different Python versions
 
 # Copyright (C) 2021 The Psycopg Team
 
+from __future__ import annotations
+
 import sys
-import asyncio
-from typing import Any, Awaitable, Generator, Optional, Union, Type, TypeVar
-from typing_extensions import TypeAlias
 
 import psycopg.errors as e
 
-T = TypeVar("T")
-FutureT: TypeAlias = Union["asyncio.Future[T]", Generator[Any, None, T], Awaitable[T]]
-
-if sys.version_info >= (3, 8):
-    create_task = asyncio.create_task
-    Task = asyncio.Task
-
-else:
-
-    def create_task(
-        coro: FutureT[T], name: Optional[str] = None
-    ) -> "asyncio.Future[T]":
-        return asyncio.create_task(coro)
-
-    Task = asyncio.Future
-
 if sys.version_info >= (3, 9):
     from collections import Counter, deque as Deque
 else:
     from typing import Counter, Deque
 
+if sys.version_info >= (3, 10):
+    from typing import TypeAlias
+else:
+    from typing_extensions import TypeAlias
+
 if sys.version_info >= (3, 11):
     from typing import Self
 else:
     from typing_extensions import Self
 
+if sys.version_info >= (3, 13):
+    from typing import TypeVar
+else:
+    from typing_extensions import TypeVar
+
 __all__ = [
     "Counter",
     "Deque",
     "Self",
-    "Task",
-    "create_task",
+    "TypeAlias",
+    "TypeVar",
 ]
 
 # Workaround for psycopg < 3.0.8.
 # Timeout on NullPool connection mignt not work correctly.
 try:
-    ConnectionTimeout: Type[e.OperationalError] = e.ConnectionTimeout
+    ConnectionTimeout: type[e.OperationalError] = e.ConnectionTimeout
 except AttributeError:
 
     class DummyConnectionTimeout(e.OperationalError):
