@@ -7,6 +7,7 @@ Types used in the psycopg_pool package
 from __future__ import annotations
 
 from typing import Awaitable, Callable, TYPE_CHECKING
+from typing import Union  # drop with Python 3.8
 
 from ._compat import TypeAlias, TypeVar
 
@@ -26,8 +27,8 @@ ConnectionCB: TypeAlias = Callable[[CT], None]
 AsyncConnectionCB: TypeAlias = Callable[[ACT], Awaitable[None]]
 
 # Callbacks to pass the pool to on connection failure
-ConnectFailedCB: TypeAlias = "Callable[[ConnectionPool[Any]], None]"
-AsyncConnectFailedCB: TypeAlias = (
-    "Callable[[AsyncConnectionPool[Any]], None]"
-    "| Callable[[AsyncConnectionPool[Any]], Awaitable[None]]"
-)
+ConnectFailedCB: TypeAlias = Callable[["ConnectionPool[Any]"], None]
+AsyncConnectFailedCB: TypeAlias = Union[
+    Callable[["AsyncConnectionPool[Any]"], None],
+    Callable[["AsyncConnectionPool[Any]"], Awaitable[None]],
+]
