@@ -1,7 +1,9 @@
+from __future__ import annotations
+
 import os
 import sys
 import ctypes
-from typing import Iterator, List, NamedTuple
+from typing import Iterator, NamedTuple
 from tempfile import TemporaryFile
 
 import pytest
@@ -90,7 +92,7 @@ def trace(libpq):
 
 class Tracer:
     def trace(self, conn):
-        pgconn: "pq.abc.PGconn"
+        pgconn: pq.abc.PGconn
 
         if hasattr(conn, "exec_"):
             pgconn = conn
@@ -103,7 +105,7 @@ class Tracer:
 
 
 class TraceLog:
-    def __init__(self, pgconn: "pq.abc.PGconn"):
+    def __init__(self, pgconn: pq.abc.PGconn):
         self.pgconn = pgconn
         self.tempfile = TemporaryFile(buffering=0)
         pgconn.trace(self.tempfile.fileno())
@@ -114,13 +116,12 @@ class TraceLog:
             self.pgconn.untrace()
         self.tempfile.close()
 
-    def __iter__(self) -> "Iterator[TraceEntry]":
+    def __iter__(self) -> Iterator[TraceEntry]:
         self.tempfile.seek(0)
         data = self.tempfile.read()
-        for entry in self._parse_entries(data):
-            yield entry
+        yield from self._parse_entries(data)
 
-    def _parse_entries(self, data: bytes) -> "Iterator[TraceEntry]":
+    def _parse_entries(self, data: bytes) -> Iterator[TraceEntry]:
         for line in data.splitlines():
             direction, length, type, *content = line.split(b"\t")
             yield TraceEntry(
@@ -138,4 +139,4 @@ class TraceEntry(NamedTuple):
     direction: str
     length: int
     type: str
-    content: List[bytes]
+    content: list[bytes]
