@@ -1,13 +1,19 @@
-import asyncio
+from __future__ import annotations
+
 import logging
+import weakref
 from time import time
-from typing import Any, List, Tuple
+from typing import Any, Dict
 
 import pytest
 
 import psycopg
 from psycopg.pq import TransactionStatus
-from psycopg._compat import create_task, Counter
+from psycopg.rows import class_row, Row, TupleRow
+
+from ..utils import assert_type, Counter, set_autocommit
+from ..acompat import AEvent, spawn, gather, asleep, skip_sync
+from .test_pool_common_async import delay_connection
 
 try:
     import psycopg_pool as pool
@@ -15,16 +21,13 @@ except ImportError:
     # Tests should have been skipped if the package is not available
     pass
 
-pytestmark = [pytest.mark.anyio]
+if True:  # ASYNC
+    pytestmark = [pytest.mark.anyio]
 
 
-async def test_defaults(dsn):
+async def test_default_sizes(dsn):
     async with pool.AsyncConnectionPool(dsn) as p:
         assert p.min_size == p.max_size == 4
-        assert p.timeout == 30
-        assert p.max_idle == 10 * 60
-        assert p.max_lifetime == 60 * 60
-        assert p.num_workers == 3
 
 
 @pytest.mark.parametrize("min_size, max_size", [(2, None), (0, 2), (2, 4)])
@@ -40,21 +43,66 @@ async def test_bad_size(dsn, min_size, max_size):
         pool.AsyncConnectionPool(min_size=min_size, max_size=max_size)
 
 
-async def test_connection_class(dsn):
-    class MyConn(psycopg.AsyncConnection[Any]):
+class MyRow(Dict[str, Any]):
+    pass
+
+
+async def test_generic_connection_type(dsn):
+    async def configure(conn: psycopg.AsyncConnection[Any]) -> None:
+        await set_autocommit(conn, True)
+
+    class MyConnection(psycopg.AsyncConnection[Row]):
         pass
 
-    async with pool.AsyncConnectionPool(dsn, connection_class=MyConn, min_size=1) as p:
-        async with p.connection() as conn:
-            assert isinstance(conn, MyConn)
-
-
-async def test_kwargs(dsn):
     async with pool.AsyncConnectionPool(
-        dsn, kwargs={"autocommit": True}, min_size=1
-    ) as p:
-        async with p.connection() as conn:
-            assert conn.autocommit
+        dsn,
+        connection_class=MyConnection[MyRow],
+        kwargs=dict(row_factory=class_row(MyRow)),
+        configure=configure,
+    ) as p1:
+        async with p1.connection() as conn1:
+            cur1 = await conn1.execute("select 1 as x")
+            (row1,) = await cur1.fetchall()
+    assert_type(p1, pool.AsyncConnectionPool[MyConnection[MyRow]])
+    assert_type(conn1, MyConnection[MyRow])
+    assert_type(row1, MyRow)
+    assert conn1.autocommit
+    assert row1 == {"x": 1}
+
+    async with pool.AsyncConnectionPool(
+        dsn, connection_class=MyConnection[TupleRow]
+    ) as p2:
+        async with p2.connection() as conn2:
+            cur2 = await conn2.execute("select 2 as y")
+            (row2,) = await cur2.fetchall()
+    assert_type(p2, pool.AsyncConnectionPool[MyConnection[TupleRow]])
+    assert_type(conn2, MyConnection[TupleRow])
+    assert_type(row2, TupleRow)
+    assert row2 == (2,)
+
+
+async def test_non_generic_connection_type(dsn):
+    async def configure(conn: psycopg.AsyncConnection[Any]) -> None:
+        await set_autocommit(conn, True)
+
+    class MyConnection(psycopg.AsyncConnection[MyRow]):
+        def __init__(self, *args: Any, **kwargs: Any):
+            kwargs["row_factory"] = class_row(MyRow)
+            super().__init__(*args, **kwargs)
+
+    async with pool.AsyncConnectionPool(
+        dsn,
+        connection_class=MyConnection,
+        configure=configure,
+    ) as p1:
+        async with p1.connection() as conn1:
+            cur1 = await conn1.execute("select 1 as x")
+            (row1,) = await cur1.fetchall()
+    assert_type(p1, pool.AsyncConnectionPool[MyConnection])
+    assert_type(conn1, MyConnection)
+    assert_type(row1, MyRow)
+    assert conn1.autocommit
+    assert row1 == {"x": 1}
 
 
 @pytest.mark.crdb_skip("backend pid")
@@ -70,12 +118,6 @@ async def test_its_really_a_pool(dsn):
             assert conn.info.backend_pid in (pid1, pid2)
 
 
-async def test_context(dsn):
-    async with pool.AsyncConnectionPool(dsn, min_size=1) as p:
-        assert not p.closed
-    assert p.closed
-
-
 @pytest.mark.crdb_skip("backend pid")
 async def test_connection_not_lost(dsn):
     async with pool.AsyncConnectionPool(dsn, min_size=1) as p:
@@ -100,7 +142,7 @@ async def test_concurrent_filling(dsn, monkeypatch):
     add_orig = pool.AsyncConnectionPool._add_to_pool
     monkeypatch.setattr(pool.AsyncConnectionPool, "_add_to_pool", add_time)
 
-    times: List[float] = []
+    times: list[float] = []
     t0 = time()
 
     async with pool.AsyncConnectionPool(dsn, min_size=5, num_workers=2) as p:
@@ -127,33 +169,6 @@ async def test_wait_ready(dsn, monkeypatch):
         await p.wait(0.0001)  # idempotent
 
 
-async def test_wait_closed(dsn):
-    async with pool.AsyncConnectionPool(dsn) as p:
-        pass
-
-    with pytest.raises(pool.PoolClosed):
-        await p.wait()
-
-
-@pytest.mark.slow
-async def test_setup_no_timeout(dsn, proxy):
-    with pytest.raises(pool.PoolTimeout):
-        async with pool.AsyncConnectionPool(
-            proxy.client_dsn, min_size=1, num_workers=1
-        ) as p:
-            await p.wait(0.2)
-
-    async with pool.AsyncConnectionPool(
-        proxy.client_dsn, min_size=1, num_workers=1
-    ) as p:
-        await asyncio.sleep(0.5)
-        assert not p._pool
-        proxy.start()
-
-        async with p.connection() as conn:
-            await conn.execute("select 1")
-
-
 async def test_configure(dsn):
     inits = 0
 
@@ -168,49 +183,18 @@ async def test_configure(dsn):
         async with p.connection() as conn:
             assert inits == 1
             res = await conn.execute("show default_transaction_read_only")
-            assert (await res.fetchone())[0] == "on"  # type: ignore[index]
+            assert (await res.fetchone())[0] == "on"
 
         async with p.connection() as conn:
             assert inits == 1
             res = await conn.execute("show default_transaction_read_only")
-            assert (await res.fetchone())[0] == "on"  # type: ignore[index]
+            assert (await res.fetchone())[0] == "on"
             await conn.close()
 
         async with p.connection() as conn:
             assert inits == 2
             res = await conn.execute("show default_transaction_read_only")
-            assert (await res.fetchone())[0] == "on"  # type: ignore[index]
-
-
-@pytest.mark.slow
-async def test_configure_badstate(dsn, caplog):
-    caplog.set_level(logging.WARNING, logger="psycopg.pool")
-
-    async def configure(conn):
-        await conn.execute("select 1")
-
-    async with pool.AsyncConnectionPool(dsn, min_size=1, configure=configure) as p:
-        with pytest.raises(pool.PoolTimeout):
-            await p.wait(timeout=0.5)
-
-    assert caplog.records
-    assert "INTRANS" in caplog.records[0].message
-
-
-@pytest.mark.slow
-async def test_configure_broken(dsn, caplog):
-    caplog.set_level(logging.WARNING, logger="psycopg.pool")
-
-    async def configure(conn):
-        async with conn.transaction():
-            await conn.execute("WAT")
-
-    async with pool.AsyncConnectionPool(dsn, min_size=1, configure=configure) as p:
-        with pytest.raises(pool.PoolTimeout):
-            await p.wait(timeout=0.5)
-
-    assert caplog.records
-    assert "WAT" in caplog.records[0].message
+            assert (await res.fetchone())[0] == "on"
 
 
 async def test_reset(dsn):
@@ -285,164 +269,6 @@ async def test_reset_broken(dsn, caplog):
     assert "WAT" in caplog.records[0].message
 
 
-@pytest.mark.slow
-@pytest.mark.timing
-@pytest.mark.crdb_skip("backend pid")
-async def test_queue(dsn):
-    async def worker(n):
-        t0 = time()
-        async with p.connection() as conn:
-            await conn.execute("select pg_sleep(0.2)")
-            pid = conn.info.backend_pid
-        t1 = time()
-        results.append((n, t1 - t0, pid))
-
-    results: List[Tuple[int, float, int]] = []
-    async with pool.AsyncConnectionPool(dsn, min_size=2) as p:
-        await p.wait()
-        ts = [create_task(worker(i)) for i in range(6)]
-        await asyncio.gather(*ts)
-
-    times = [item[1] for item in results]
-    want_times = [0.2, 0.2, 0.4, 0.4, 0.6, 0.6]
-    for got, want in zip(times, want_times):
-        assert got == pytest.approx(want, 0.1), times
-
-    assert len(set(r[2] for r in results)) == 2, results
-
-
-@pytest.mark.slow
-async def test_queue_size(dsn):
-    async def worker(t, ev=None):
-        try:
-            async with p.connection():
-                if ev:
-                    ev.set()
-                await asyncio.sleep(t)
-        except pool.TooManyRequests as e:
-            errors.append(e)
-        else:
-            success.append(True)
-
-    errors: List[Exception] = []
-    success: List[bool] = []
-
-    async with pool.AsyncConnectionPool(dsn, min_size=1, max_waiting=3) as p:
-        await p.wait()
-        ev = asyncio.Event()
-        create_task(worker(0.3, ev))
-        await ev.wait()
-
-        ts = [create_task(worker(0.1)) for i in range(4)]
-        await asyncio.gather(*ts)
-
-    assert len(success) == 4
-    assert len(errors) == 1
-    assert isinstance(errors[0], pool.TooManyRequests)
-    assert p.name in str(errors[0])
-    assert str(p.max_waiting) in str(errors[0])
-    assert p.get_stats()["requests_errors"] == 1
-
-
-@pytest.mark.slow
-@pytest.mark.timing
-@pytest.mark.crdb_skip("backend pid")
-async def test_queue_timeout(dsn):
-    async def worker(n):
-        t0 = time()
-        try:
-            async with p.connection() as conn:
-                await conn.execute("select pg_sleep(0.2)")
-                pid = conn.info.backend_pid
-        except pool.PoolTimeout as e:
-            t1 = time()
-            errors.append((n, t1 - t0, e))
-        else:
-            t1 = time()
-            results.append((n, t1 - t0, pid))
-
-    results: List[Tuple[int, float, int]] = []
-    errors: List[Tuple[int, float, Exception]] = []
-
-    async with pool.AsyncConnectionPool(dsn, min_size=2, timeout=0.1) as p:
-        ts = [create_task(worker(i)) for i in range(4)]
-        await asyncio.gather(*ts)
-
-    assert len(results) == 2
-    assert len(errors) == 2
-    for e in errors:
-        assert 0.1 < e[1] < 0.15
-
-
-@pytest.mark.slow
-@pytest.mark.timing
-async def test_dead_client(dsn):
-    async def worker(i, timeout):
-        try:
-            async with p.connection(timeout=timeout) as conn:
-                await conn.execute("select pg_sleep(0.3)")
-                results.append(i)
-        except pool.PoolTimeout:
-            if timeout > 0.2:
-                raise
-
-    async with pool.AsyncConnectionPool(dsn, min_size=2) as p:
-        results: List[int] = []
-        ts = [
-            create_task(worker(i, timeout))
-            for i, timeout in enumerate([0.4, 0.4, 0.1, 0.4, 0.4])
-        ]
-        await asyncio.gather(*ts)
-
-        await asyncio.sleep(0.2)
-        assert set(results) == set([0, 1, 3, 4])
-        assert len(p._pool) == 2  # no connection was lost
-
-
-@pytest.mark.slow
-@pytest.mark.timing
-@pytest.mark.crdb_skip("backend pid")
-async def test_queue_timeout_override(dsn):
-    async def worker(n):
-        t0 = time()
-        timeout = 0.25 if n == 3 else None
-        try:
-            async with p.connection(timeout=timeout) as conn:
-                await conn.execute("select pg_sleep(0.2)")
-                pid = conn.info.backend_pid
-        except pool.PoolTimeout as e:
-            t1 = time()
-            errors.append((n, t1 - t0, e))
-        else:
-            t1 = time()
-            results.append((n, t1 - t0, pid))
-
-    results: List[Tuple[int, float, int]] = []
-    errors: List[Tuple[int, float, Exception]] = []
-
-    async with pool.AsyncConnectionPool(dsn, min_size=2, timeout=0.1) as p:
-        ts = [create_task(worker(i)) for i in range(4)]
-        await asyncio.gather(*ts)
-
-    assert len(results) == 3
-    assert len(errors) == 1
-    for e in errors:
-        assert 0.1 < e[1] < 0.15
-
-
-@pytest.mark.crdb_skip("backend pid")
-async def test_broken_reconnect(dsn):
-    async with pool.AsyncConnectionPool(dsn, min_size=1) as p:
-        async with p.connection() as conn:
-            pid1 = conn.info.backend_pid
-            await conn.close()
-
-        async with p.connection() as conn2:
-            pid2 = conn2.info.backend_pid
-
-    assert pid1 != pid2
-
-
 @pytest.mark.crdb_skip("backend pid")
 async def test_intrans_rollback(dsn, caplog):
     caplog.set_level(logging.WARNING, logger="psycopg.pool")
@@ -538,154 +364,25 @@ async def test_fail_rollback_close(dsn, caplog, monkeypatch):
     assert "BAD" in caplog.records[2].message
 
 
-async def test_close_no_tasks(dsn):
-    p = pool.AsyncConnectionPool(dsn)
-    assert p._sched_runner and not p._sched_runner.done()
-    assert p._workers
-    workers = p._workers[:]
-    for t in workers:
-        assert not t.done()
+async def test_del_no_warning(dsn, recwarn, gc_collect):
+    p = pool.AsyncConnectionPool(dsn, min_size=2, open=False)
+    await p.open()
+    async with p.connection() as conn:
+        await conn.execute("select 1")
 
-    await p.close()
-    assert p._sched_runner is None
-    assert not p._workers
-    for t in workers:
-        assert t.done()
-
-
-async def test_putconn_no_pool(aconn_cls, dsn):
-    async with pool.AsyncConnectionPool(dsn, min_size=1) as p:
-        conn = await aconn_cls.connect(dsn)
-        with pytest.raises(ValueError):
-            await p.putconn(conn)
-
-    await conn.close()
-
-
-async def test_putconn_wrong_pool(dsn):
-    async with pool.AsyncConnectionPool(dsn, min_size=1) as p1:
-        async with pool.AsyncConnectionPool(dsn, min_size=1) as p2:
-            conn = await p1.getconn()
-            with pytest.raises(ValueError):
-                await p2.putconn(conn)
-
-
-async def test_closed_getconn(dsn):
-    p = pool.AsyncConnectionPool(dsn, min_size=1)
-    assert not p.closed
-    async with p.connection():
-        pass
-
-    await p.close()
-    assert p.closed
-
-    with pytest.raises(pool.PoolClosed):
-        async with p.connection():
-            pass
+    await p.wait()
+    ref = weakref.ref(p)
+    del p
+    gc_collect()
+    assert not ref()
+    assert not recwarn, [str(w.message) for w in recwarn.list]
 
 
 async def test_closed_putconn(dsn):
-    p = pool.AsyncConnectionPool(dsn, min_size=1)
-
-    async with p.connection() as conn:
-        pass
-    assert not conn.closed
-
-    async with p.connection() as conn:
-        await p.close()
-    assert conn.closed
-
-
-async def test_closed_queue(dsn):
-    async def w1():
+    async with pool.AsyncConnectionPool(dsn, min_size=1) as p:
         async with p.connection() as conn:
-            e1.set()  # Tell w0 that w1 got a connection
-            cur = await conn.execute("select 1")
-            assert await cur.fetchone() == (1,)
-            await e2.wait()  # Wait until w0 has tested w2
-        success.append("w1")
-
-    async def w2():
-        try:
-            async with p.connection():
-                pass  # unexpected
-        except pool.PoolClosed:
-            success.append("w2")
-
-    e1 = asyncio.Event()
-    e2 = asyncio.Event()
-
-    p = pool.AsyncConnectionPool(dsn, min_size=1)
-    await p.wait()
-    success: List[str] = []
-
-    t1 = create_task(w1())
-    # Wait until w1 has received a connection
-    await e1.wait()
-
-    t2 = create_task(w2())
-    # Wait until w2 is in the queue
-    await ensure_waiting(p)
-    await p.close()
-
-    # Wait for the workers to finish
-    e2.set()
-    await asyncio.gather(t1, t2)
-    assert len(success) == 2
-
-
-async def test_open_explicit(dsn):
-    p = pool.AsyncConnectionPool(dsn, open=False)
-    assert p.closed
-    with pytest.raises(pool.PoolClosed):
-        await p.getconn()
-
-    with pytest.raises(pool.PoolClosed, match="is not open yet"):
-        async with p.connection():
             pass
-
-    await p.open()
-    try:
-        assert not p.closed
-
-        async with p.connection() as conn:
-            cur = await conn.execute("select 1")
-            assert await cur.fetchone() == (1,)
-
-    finally:
-        await p.close()
-
-    with pytest.raises(pool.PoolClosed, match="is already closed"):
-        await p.getconn()
-
-
-async def test_open_context(dsn):
-    p = pool.AsyncConnectionPool(dsn, open=False)
-    assert p.closed
-
-    async with p:
-        assert not p.closed
-
-        async with p.connection() as conn:
-            cur = await conn.execute("select 1")
-            assert await cur.fetchone() == (1,)
-
-    assert p.closed
-
-
-async def test_open_no_op(dsn):
-    p = pool.AsyncConnectionPool(dsn)
-    try:
-        assert not p.closed
-        await p.open()
-        assert not p.closed
-
-        async with p.connection() as conn:
-            cur = await conn.execute("select 1")
-            assert await cur.fetchone() == (1,)
-
-    finally:
-        await p.close()
+        assert not conn.closed
 
 
 @pytest.mark.slow
@@ -718,17 +415,6 @@ async def test_open_as_wait(dsn, monkeypatch):
         await p.open(wait=True, timeout=0.5)
 
 
-async def test_reopen(dsn):
-    p = pool.AsyncConnectionPool(dsn)
-    async with p.connection() as conn:
-        await conn.execute("select 1")
-    await p.close()
-    assert p._sched_runner is None
-
-    with pytest.raises(psycopg.OperationalError, match="cannot be reused"):
-        await p.open()
-
-
 @pytest.mark.slow
 @pytest.mark.timing
 @pytest.mark.parametrize(
@@ -752,11 +438,9 @@ async def test_grow(dsn, monkeypatch, min_size, want_times):
         dsn, min_size=min_size, max_size=4, num_workers=3
     ) as p:
         await p.wait(1.0)
-        ts = []
-        results: List[Tuple[int, float]] = []
-
-        ts = [create_task(worker(i)) for i in range(len(want_times))]
-        await asyncio.gather(*ts)
+        results: list[tuple[int, float]] = []
+        ts = [spawn(worker, args=(i,)) for i in range(len(want_times))]
+        await gather(*ts)
 
     times = [item[1] for item in results]
     for got, want in zip(times, want_times):
@@ -768,7 +452,7 @@ async def test_grow(dsn, monkeypatch, min_size, want_times):
 async def test_shrink(dsn, monkeypatch):
     from psycopg_pool.pool_async import ShrinkPool
 
-    results: List[Tuple[int, int]] = []
+    results: list[tuple[int, int]] = []
 
     async def run_hacked(self, pool):
         n0 = pool._nconns
@@ -787,10 +471,10 @@ async def test_shrink(dsn, monkeypatch):
         await p.wait(5.0)
         assert p.max_idle == 0.2
 
-        ts = [create_task(worker(i)) for i in range(4)]
-        await asyncio.gather(*ts)
+        ts = [spawn(worker, args=(i,)) for i in range(4)]
+        await gather(*ts)
 
-        await asyncio.sleep(1)
+        await asleep(1)
 
     assert results == [(4, 4), (4, 3), (3, 2), (2, 2), (2, 2)]
 
@@ -800,10 +484,10 @@ async def test_shrink(dsn, monkeypatch):
 async def test_reconnect(proxy, caplog, monkeypatch):
     caplog.set_level(logging.WARNING, logger="psycopg.pool")
 
-    assert pool.base.ConnectionAttempt.INITIAL_DELAY == 1.0
-    assert pool.base.ConnectionAttempt.DELAY_JITTER == 0.1
-    monkeypatch.setattr(pool.base.ConnectionAttempt, "INITIAL_DELAY", 0.1)
-    monkeypatch.setattr(pool.base.ConnectionAttempt, "DELAY_JITTER", 0.0)
+    assert pool.base.AttemptWithBackoff.INITIAL_DELAY == 1.0
+    assert pool.base.AttemptWithBackoff.DELAY_JITTER == 0.1
+    monkeypatch.setattr(pool.base.AttemptWithBackoff, "INITIAL_DELAY", 0.1)
+    monkeypatch.setattr(pool.base.AttemptWithBackoff, "DELAY_JITTER", 0.0)
 
     caplog.clear()
     proxy.start()
@@ -815,7 +499,7 @@ async def test_reconnect(proxy, caplog, monkeypatch):
             async with p.connection() as conn:
                 await conn.execute("select 1")
 
-        await asyncio.sleep(1.0)
+        await asleep(1.0)
         proxy.start()
         await p.wait()
 
@@ -829,21 +513,31 @@ async def test_reconnect(proxy, caplog, monkeypatch):
     assert len(deltas) == 3
     want = 0.1
     for delta in deltas:
-        assert delta == pytest.approx(want, 0.05), deltas
+        assert delta == pytest.approx(want, 0.1), deltas
         want *= 2
 
 
 @pytest.mark.slow
 @pytest.mark.timing
-async def test_reconnect_failure(proxy):
+@pytest.mark.parametrize("async_cb", [pytest.param(True, marks=skip_sync), False])
+async def test_reconnect_failure(proxy, async_cb):
     proxy.start()
 
     t1 = None
 
-    def failed(pool):
-        assert pool.name == "this-one"
-        nonlocal t1
-        t1 = time()
+    if async_cb:
+
+        async def failed(pool):
+            assert pool.name == "this-one"
+            nonlocal t1
+            t1 = time()
+
+    else:
+
+        def failed(pool):
+            assert pool.name == "this-one"
+            nonlocal t1
+            t1 = time()
 
     async with pool.AsyncConnectionPool(
         proxy.client_dsn,
@@ -860,7 +554,7 @@ async def test_reconnect_failure(proxy):
                 await conn.execute("select 1")
 
         t0 = time()
-        await asyncio.sleep(1.5)
+        await asleep(1.5)
         assert t1
         assert t1 - t0 == pytest.approx(1.0, 0.1)
         assert p._nconns == 0
@@ -879,7 +573,7 @@ async def test_reconnect_after_grow_failed(proxy):
     # in grow mode. See issue #370.
     proxy.stop()
 
-    ev = asyncio.Event()
+    ev = AEvent()
 
     def failed(pool):
         ev.set()
@@ -887,14 +581,14 @@ async def test_reconnect_after_grow_failed(proxy):
     async with pool.AsyncConnectionPool(
         proxy.client_dsn, min_size=4, reconnect_timeout=1.0, reconnect_failed=failed
     ) as p:
-        await asyncio.wait_for(ev.wait(), 2.0)
+        await ev.wait_timeout(2.0)
 
         with pytest.raises(pool.PoolTimeout):
             async with p.connection(timeout=0.5) as conn:
                 pass
 
         ev.clear()
-        await asyncio.wait_for(ev.wait(), 2.0)
+        await ev.wait_timeout(2.0)
 
         proxy.start()
 
@@ -908,7 +602,7 @@ async def test_reconnect_after_grow_failed(proxy):
 @pytest.mark.slow
 async def test_refill_on_check(proxy):
     proxy.start()
-    ev = asyncio.Event()
+    ev = AEvent()
 
     def failed(pool):
         ev.set()
@@ -924,7 +618,7 @@ async def test_refill_on_check(proxy):
 
         # Checking the pool will empty it
         await p.check()
-        await asyncio.wait_for(ev.wait(), 2.0)
+        await ev.wait_timeout(2.0)
         assert len(p._pool) == 0
 
         # Allow to connect again
@@ -942,20 +636,20 @@ async def test_uniform_use(dsn):
         counts = Counter[int]()
         for i in range(8):
             async with p.connection() as conn:
-                await asyncio.sleep(0.1)
+                await asleep(0.1)
                 counts[id(conn)] += 1
 
     assert len(counts) == 4
-    assert set(counts.values()) == set([2])
+    assert set(counts.values()) == {2}
 
 
 @pytest.mark.slow
 @pytest.mark.timing
 async def test_resize(dsn):
     async def sampler():
-        await asyncio.sleep(0.05)  # ensure sampling happens after shrink check
+        await asleep(0.05)  # ensure sampling happens after shrink check
         while True:
-            await asyncio.sleep(0.2)
+            await asleep(0.2)
             if p.closed:
                 break
             size.append(len(p._pool))
@@ -964,28 +658,28 @@ async def test_resize(dsn):
         async with p.connection() as conn:
             await conn.execute("select pg_sleep(%s)", [t])
 
-    size: List[int] = []
+    size: list[int] = []
 
     async with pool.AsyncConnectionPool(dsn, min_size=2, max_idle=0.2) as p:
-        s = create_task(sampler())
+        s = spawn(sampler)
 
-        await asyncio.sleep(0.3)
+        await asleep(0.3)
 
-        c = create_task(client(0.4))
+        c = spawn(client, args=(0.4,))
 
-        await asyncio.sleep(0.2)
+        await asleep(0.2)
         await p.resize(4)
         assert p.min_size == 4
         assert p.max_size == 4
 
-        await asyncio.sleep(0.4)
+        await asleep(0.4)
         await p.resize(2)
         assert p.min_size == 2
         assert p.max_size == 2
 
-        await asyncio.sleep(0.6)
+        await asleep(0.6)
 
-    await asyncio.gather(s, c)
+    await gather(s, c)
     assert size == [2, 1, 3, 4, 3, 2, 2]
 
 
@@ -996,23 +690,17 @@ async def test_bad_resize(dsn, min_size, max_size):
             await p.resize(min_size=min_size, max_size=max_size)
 
 
-async def test_jitter():
-    rnds = [pool.AsyncConnectionPool._jitter(30, -0.1, +0.2) for i in range(100)]
-    assert 27 <= min(rnds) <= 28
-    assert 35 < max(rnds) < 36
-
-
 @pytest.mark.slow
 @pytest.mark.timing
 @pytest.mark.crdb_skip("backend pid")
 async def test_max_lifetime(dsn):
     async with pool.AsyncConnectionPool(dsn, min_size=1, max_lifetime=0.2) as p:
-        await asyncio.sleep(0.1)
+        await asleep(0.1)
         pids = []
         for i in range(5):
             async with p.connection() as conn:
                 pids.append(conn.info.backend_pid)
-            await asyncio.sleep(0.2)
+            await asleep(0.2)
 
     assert pids[0] == pids[1] != pids[4], pids
 
@@ -1026,7 +714,7 @@ async def test_check(dsn, caplog):
             pid = conn.info.backend_pid
 
         await p.wait(1.0)
-        pids = set(conn.info.backend_pid for conn in p._pool)
+        pids = {conn.info.backend_pid for conn in p._pool}
         assert pid in pids
         await conn.close()
 
@@ -1034,7 +722,7 @@ async def test_check(dsn, caplog):
         await p.check()
         assert len(caplog.records) == 1
         await p.wait(1.0)
-        pids2 = set(conn.info.backend_pid for conn in p._pool)
+        pids2 = {conn.info.backend_pid for conn in p._pool}
         assert len(pids & pids2) == 3
         assert pid not in pids2
 
@@ -1047,6 +735,116 @@ async def test_check_idle(dsn):
             assert conn.info.transaction_status == TransactionStatus.IDLE
 
 
+@pytest.mark.crdb_skip("pg_terminate_backend")
+async def test_connect_no_check(dsn):
+    async with pool.AsyncConnectionPool(dsn, min_size=2) as p:
+        await p.wait(1.0)
+        async with p.connection() as conn:
+            async with p.connection() as conn2:
+                pid2 = conn2.info.backend_pid
+            await conn.execute("select pg_terminate_backend(%s)", [pid2])
+
+        with pytest.raises(psycopg.OperationalError):
+            async with p.connection() as conn:
+                await conn.execute("select 1")
+                async with p.connection() as conn2:
+                    await conn2.execute("select 2")
+
+
+@pytest.mark.crdb_skip("pg_terminate_backend")
+@pytest.mark.parametrize("autocommit", [True, False])
+async def test_connect_check(dsn, caplog, autocommit):
+    caplog.set_level(logging.WARNING, logger="psycopg.pool")
+
+    async with pool.AsyncConnectionPool(
+        dsn,
+        min_size=2,
+        kwargs={"autocommit": autocommit},
+        check=pool.AsyncConnectionPool.check_connection,
+    ) as p:
+        await p.wait(1.0)
+        async with p.connection() as conn:
+            pid1 = conn.info.backend_pid
+            async with p.connection() as conn2:
+                pid2 = conn2.info.backend_pid
+            await conn.execute("select pg_terminate_backend(%s)", [pid2])
+
+        async with p.connection() as conn:
+            assert conn.info.transaction_status == TransactionStatus.IDLE
+            await conn.execute("select 1")
+            async with p.connection() as conn2:
+                assert conn2.info.transaction_status == TransactionStatus.IDLE
+                await conn2.execute("select 2")
+
+                pids = {c.info.backend_pid for c in [conn, conn2]}
+
+    assert pid1 in pids
+    assert pid2 not in pids
+    assert not caplog.records
+
+
+@pytest.mark.parametrize("autocommit", [True, False])
+@pytest.mark.crdb_skip("pg_terminate_backend")
+async def test_getconn_check(dsn, caplog, autocommit):
+    caplog.set_level(logging.WARNING, logger="psycopg.pool")
+
+    async with pool.AsyncConnectionPool(
+        dsn,
+        kwargs={"autocommit": autocommit},
+        min_size=2,
+        check=pool.AsyncConnectionPool.check_connection,
+    ) as p:
+        await p.wait(1.0)
+        async with p.connection() as conn:
+            pid1 = conn.info.backend_pid
+            async with p.connection() as conn2:
+                pid2 = conn2.info.backend_pid
+            await conn.execute("select pg_terminate_backend(%s)", [pid2])
+
+        conn = await p.getconn()
+        try:
+            assert conn.info.transaction_status == TransactionStatus.IDLE
+            await conn.execute("select 1")
+            await conn.rollback()
+            conn2 = await p.getconn()
+            try:
+                assert conn2.info.transaction_status == TransactionStatus.IDLE
+                await conn2.execute("select 1")
+                await conn2.rollback()
+                pids = {c.info.backend_pid for c in [conn, conn2]}
+            finally:
+                await p.putconn(conn2)
+        finally:
+            await p.putconn(conn)
+
+    assert pid1 in pids
+    assert pid2 not in pids
+    assert not caplog.records
+
+
+@pytest.mark.slow
+async def test_connect_check_timeout(dsn, proxy):
+    proxy.start()
+    async with pool.AsyncConnectionPool(
+        proxy.client_dsn,
+        min_size=1,
+        timeout=1.0,
+        check=pool.AsyncConnectionPool.check_connection,
+    ) as p:
+        await p.wait()
+
+        proxy.stop()
+        t0 = time()
+        with pytest.raises(pool.PoolTimeout):
+            async with p.connection():
+                pass
+        assert 1.0 <= (time() - t0) <= 1.1
+
+        proxy.start()
+        async with p.connection(timeout=10) as conn:
+            await conn.execute("select 1")
+
+
 @pytest.mark.slow
 async def test_check_max_lifetime(dsn):
     async with pool.AsyncConnectionPool(dsn, min_size=1, max_lifetime=0.2) as p:
@@ -1054,87 +852,14 @@ async def test_check_max_lifetime(dsn):
             pid = conn.info.backend_pid
         async with p.connection() as conn:
             assert conn.info.backend_pid == pid
-        await asyncio.sleep(0.3)
+        await asleep(0.3)
         await p.check()
         async with p.connection() as conn:
             assert conn.info.backend_pid != pid
 
 
 @pytest.mark.slow
-@pytest.mark.timing
-async def test_stats_measures(dsn):
-    async def worker(n):
-        async with p.connection() as conn:
-            await conn.execute("select pg_sleep(0.2)")
-
-    async with pool.AsyncConnectionPool(dsn, min_size=2, max_size=4) as p:
-        await p.wait(2.0)
-
-        stats = p.get_stats()
-        assert stats["pool_min"] == 2
-        assert stats["pool_max"] == 4
-        assert stats["pool_size"] == 2
-        assert stats["pool_available"] == 2
-        assert stats["requests_waiting"] == 0
-
-        ts = [create_task(worker(i)) for i in range(3)]
-        await asyncio.sleep(0.1)
-        stats = p.get_stats()
-        await asyncio.gather(*ts)
-        assert stats["pool_min"] == 2
-        assert stats["pool_max"] == 4
-        assert stats["pool_size"] == 3
-        assert stats["pool_available"] == 0
-        assert stats["requests_waiting"] == 0
-
-        await p.wait(2.0)
-        ts = [create_task(worker(i)) for i in range(7)]
-        await asyncio.sleep(0.1)
-        stats = p.get_stats()
-        await asyncio.gather(*ts)
-        assert stats["pool_min"] == 2
-        assert stats["pool_max"] == 4
-        assert stats["pool_size"] == 4
-        assert stats["pool_available"] == 0
-        assert stats["requests_waiting"] == 3
-
-
-@pytest.mark.slow
-@pytest.mark.timing
-async def test_stats_usage(dsn):
-    async def worker(n):
-        try:
-            async with p.connection(timeout=0.3) as conn:
-                await conn.execute("select pg_sleep(0.2)")
-        except pool.PoolTimeout:
-            pass
-
-    async with pool.AsyncConnectionPool(dsn, min_size=3) as p:
-        await p.wait(2.0)
-
-        ts = [create_task(worker(i)) for i in range(7)]
-        await asyncio.gather(*ts)
-        stats = p.get_stats()
-        assert stats["requests_num"] == 7
-        assert stats["requests_queued"] == 4
-        assert 850 <= stats["requests_wait_ms"] <= 950
-        assert stats["requests_errors"] == 1
-        assert 1150 <= stats["usage_ms"] <= 1250
-        assert stats.get("returns_bad", 0) == 0
-
-        async with p.connection() as conn:
-            await conn.close()
-        await p.wait()
-        stats = p.pop_stats()
-        assert stats["requests_num"] == 8
-        assert stats["returns_bad"] == 1
-        async with p.connection():
-            pass
-        assert p.get_stats()["requests_num"] == 1
-
-
-@pytest.mark.slow
-async def test_stats_connect(dsn, proxy, monkeypatch):
+async def test_stats_connect(proxy, monkeypatch):
     proxy.start()
     delay_connection(monkeypatch, 0.2)
     async with pool.AsyncConnectionPool(proxy.client_dsn, min_size=3) as p:
@@ -1147,13 +872,32 @@ async def test_stats_connect(dsn, proxy, monkeypatch):
 
         proxy.stop()
         await p.check()
-        await asyncio.sleep(0.1)
+        await asleep(0.1)
         stats = p.get_stats()
         assert stats["connections_num"] > 3
         assert stats["connections_errors"] > 0
         assert stats["connections_lost"] == 3
 
 
+@pytest.mark.crdb_skip("pg_terminate_backend")
+async def test_stats_check(dsn):
+    async with pool.AsyncConnectionPool(
+        dsn, min_size=1, check=pool.AsyncConnectionPool.check_connection
+    ) as p:
+        await p.wait()
+        async with p.connection() as conn:
+            pid = conn.info.backend_pid
+
+        async with await psycopg.AsyncConnection.connect(dsn) as conn:
+            await conn.execute("select pg_terminate_backend(%s)", [pid])
+
+        async with p.connection() as conn:
+            assert conn.info.backend_pid != pid
+
+        stats = p.get_stats()
+        assert stats["connections_lost"] == 1
+
+
 @pytest.mark.slow
 async def test_spike(dsn, monkeypatch):
     # Inspired to https://github.com/brettwooldridge/HikariCP/blob/dev/
@@ -1162,13 +906,13 @@ async def test_spike(dsn, monkeypatch):
 
     async def worker():
         async with p.connection():
-            await asyncio.sleep(0.002)
+            await asleep(0.002)
 
     async with pool.AsyncConnectionPool(dsn, min_size=5, max_size=10) as p:
         await p.wait()
 
-        ts = [create_task(worker()) for i in range(50)]
-        await asyncio.gather(*ts)
+        ts = [spawn(worker) for i in range(50)]
+        await gather(*ts)
         await p.wait()
 
         assert len(p._pool) < 7
@@ -1183,14 +927,14 @@ async def test_debug_deadlock(dsn):
     handler.setLevel(logging.DEBUG)
     logger.addHandler(handler)
     try:
-        async with pool.AsyncConnectionPool(dsn, min_size=4, open=True) as p:
+        async with pool.AsyncConnectionPool(dsn, min_size=4) as p:
             await p.wait(timeout=2)
     finally:
         logger.removeHandler(handler)
         logger.setLevel(old_level)
 
 
-@pytest.mark.skipif("sys.version_info < (3, 8)", reason="asyncio bug")
+@skip_sync
 async def test_cancellation_in_queue(dsn):
     # https://github.com/psycopg/psycopg/issues/509
 
@@ -1200,7 +944,7 @@ async def test_cancellation_in_queue(dsn):
         await p.wait()
 
         got_conns = []
-        ev = asyncio.Event()
+        ev = AEvent()
 
         async def worker(i):
             try:
@@ -1216,28 +960,27 @@ async def test_cancellation_in_queue(dsn):
                     if len(got_conns) >= nconns:
                         ev.set()
 
-                    await asyncio.sleep(5)
+                    await asleep(5)
 
             except BaseException as ex:
                 logging.info("worker %s stopped: %r", i, ex)
                 raise
 
         # Start tasks taking up all the connections and getting in the queue
-        tasks = [asyncio.ensure_future(worker(i)) for i in range(nconns * 3)]
+        tasks = [spawn(worker, (i,)) for i in range(nconns * 3)]
 
         # wait until the pool has served all the connections and clients are queued.
-        await asyncio.wait_for(ev.wait(), 3.0)
+        await ev.wait_timeout(3.0)
         for i in range(10):
             if p.get_stats().get("requests_queued", 0):
                 break
             else:
-                await asyncio.sleep(0.1)
+                await asleep(0.1)
         else:
             pytest.fail("no client got in the queue")
 
         [task.cancel() for task in reversed(tasks)]
-        # Python 3.7 hangs on this statement, instead of timing out or returning
-        await asyncio.wait_for(asyncio.gather(*tasks, return_exceptions=True), 1.0)
+        await gather(*tasks, return_exceptions=True, timeout=1.0)
 
         stats = p.get_stats()
         assert stats["pool_available"] == 3
@@ -1248,22 +991,32 @@ async def test_cancellation_in_queue(dsn):
             assert await cur.fetchone() == (1,)
 
 
-def delay_connection(monkeypatch, sec):
-    """
-    Return a _connect_gen function delayed by the amount of seconds
-    """
+@pytest.mark.slow
+@pytest.mark.timing
+async def test_check_backoff(dsn, caplog, monkeypatch):
+    caplog.set_level(logging.INFO, logger="psycopg.pool")
 
-    async def connect_delay(*args, **kwargs):
-        t0 = time()
-        rv = await connect_orig(*args, **kwargs)
-        t1 = time()
-        await asyncio.sleep(max(0, sec - (t1 - t0)))
-        return rv
+    assert pool.base.AttemptWithBackoff.INITIAL_DELAY == 1.0
+    assert pool.base.AttemptWithBackoff.DELAY_JITTER == 0.1
+    monkeypatch.setattr(pool.base.AttemptWithBackoff, "INITIAL_DELAY", 0.1)
+    monkeypatch.setattr(pool.base.AttemptWithBackoff, "DELAY_JITTER", 0.0)
 
-    connect_orig = psycopg.AsyncConnection.connect
-    monkeypatch.setattr(psycopg.AsyncConnection, "connect", connect_delay)
+    async def check(conn):
+        raise Exception()
 
+    caplog.clear()
+    async with pool.AsyncConnectionPool(dsn, min_size=1, check=check) as p:
+        await p.wait(2.0)
 
-async def ensure_waiting(p, num=1):
-    while len(p._waiting) < num:
-        await asyncio.sleep(0)
+        with pytest.raises(pool.PoolTimeout):
+            async with p.connection(timeout=1.0):
+                assert False
+
+    times = [rec.created for rec in caplog.records if "failed check" in rec.message]
+    assert times[1] - times[0] < 0.05
+    deltas = [times[i + 1] - times[i] for i in range(1, len(times) - 1)]
+    assert len(deltas) == 3
+    want = 0.1
+    for delta in deltas:
+        assert delta == pytest.approx(want, 0.05), deltas
+        want *= 2
