@@ -267,35 +267,51 @@ class TestIdentifier:
         assert sql.Identifier("foo") != "foo"
         assert sql.Identifier("foo") != sql.SQL("foo")
 
-    @pytest.mark.parametrize(
-        "args, want",
-        [
-            (("foo",), '"foo"'),
-            (("foo", "bar"), '"foo"."bar"'),
-            (("fo'o", 'ba"r'), '"fo\'o"."ba""r"'),
-        ],
-    )
+    _as_string_params = [
+        (("foo",), '"foo"'),
+        (("foo", "bar"), '"foo"."bar"'),
+        (("fo'o", 'ba"r'), '"fo\'o"."ba""r"'),
+    ]
+
+    @pytest.mark.parametrize("args, want", _as_string_params)
     def test_as_string(self, conn, args, want):
         assert sql.Identifier(*args).as_string(conn) == want
 
-    @pytest.mark.parametrize(
-        "args, want, enc",
-        [
-            crdb_encoding(("foo",), '"foo"', "ascii"),
-            crdb_encoding(("foo", "bar"), '"foo"."bar"', "ascii"),
-            crdb_encoding(("fo'o", 'ba"r'), '"fo\'o"."ba""r"', "ascii"),
-            (("foo", eur), f'"foo"."{eur}"', "utf8"),
-            crdb_encoding(("foo", eur), f'"foo"."{eur}"', "latin9"),
-        ],
-    )
+    @pytest.mark.parametrize("args, want", _as_string_params)
+    def test_as_string_no_conn(self, args, want):
+        assert sql.Identifier(*args).as_string(None) == want
+        assert sql.Identifier(*args).as_string() == want
+
+    _as_bytes_params = [
+        crdb_encoding(("foo",), '"foo"', "ascii"),
+        crdb_encoding(("foo", "bar"), '"foo"."bar"', "ascii"),
+        crdb_encoding(("fo'o", 'ba"r'), '"fo\'o"."ba""r"', "ascii"),
+        (("foo", eur), f'"foo"."{eur}"', "utf8"),
+        crdb_encoding(("foo", eur), f'"foo"."{eur}"', "latin9"),
+    ]
+
+    @pytest.mark.parametrize("args, want, enc", _as_bytes_params)
     def test_as_bytes(self, conn, args, want, enc):
         want = want.encode(enc)
         conn.execute(f"set client_encoding to {py2pgenc(enc).decode()}")
         assert sql.Identifier(*args).as_bytes(conn) == want
 
+    @pytest.mark.parametrize("args, want, enc", _as_bytes_params)
+    def test_as_bytes_no_conn(self, conn, args, want, enc):
+        want = want.encode()
+        assert sql.Identifier(*args).as_bytes(None) == want
+        assert sql.Identifier(*args).as_bytes() == want
+
     def test_join(self):
         assert not hasattr(sql.Identifier("foo"), "join")
 
+    def test_escape_no_conn(self, conn):
+        conn.execute("set client_encoding to 'utf8'")
+        for c in range(1, 128):
+            s = chr(c)
+            want = sql.Identifier(s).as_bytes(conn)
+            assert want == sql.Identifier(s).as_bytes(None)
+
 
 class TestLiteral:
     def test_class(self):
@@ -312,17 +328,40 @@ class TestLiteral:
         assert repr(sql.Literal("foo")) == "Literal('foo')"
         assert str(sql.Literal("foo")) == "Literal('foo')"
 
-    def test_as_string(self, conn):
-        assert sql.Literal(None).as_string(conn) == "NULL"
-        assert no_e(sql.Literal("foo").as_string(conn)) == "'foo'"
-        assert sql.Literal(42).as_string(conn) == "42"
-        assert sql.Literal(dt.date(2017, 1, 1)).as_string(conn) == "'2017-01-01'::date"
+    _params = [
+        (None, "NULL"),
+        ("foo", "'foo'"),
+        (42, "42"),
+        (dt.date(2017, 1, 1), "'2017-01-01'::date"),
+    ]
 
-    def test_as_bytes(self, conn):
-        assert sql.Literal(None).as_bytes(conn) == b"NULL"
-        assert no_e(sql.Literal("foo").as_bytes(conn)) == b"'foo'"
-        assert sql.Literal(42).as_bytes(conn) == b"42"
-        assert sql.Literal(dt.date(2017, 1, 1)).as_bytes(conn) == b"'2017-01-01'::date"
+    @pytest.mark.parametrize("obj, want", _params)
+    def test_as_string(self, conn, obj, want):
+        got = sql.Literal(obj).as_string(conn)
+        if isinstance(obj, str):
+            got = no_e(got)
+        assert got == want
+
+    @pytest.mark.parametrize("obj, want", _params)
+    def test_as_bytes(self, conn, obj, want):
+        got = sql.Literal(obj).as_bytes(conn)
+        if isinstance(obj, str):
+            got = no_e(got)
+        assert got == want.encode()
+
+    @pytest.mark.parametrize("obj, want", _params)
+    def test_as_string_no_conn(self, obj, want):
+        got = sql.Literal(obj).as_string()
+        if isinstance(obj, str):
+            got = no_e(got)
+        assert got == want
+
+    @pytest.mark.parametrize("obj, want", _params)
+    def test_as_bytes_no_conn(self, obj, want):
+        got = sql.Literal(obj).as_bytes()
+        if isinstance(obj, str):
+            got = no_e(got)
+        assert got == want.encode()
 
     @pytest.mark.parametrize("encoding", ["utf8", crdb_encoding("latin9")])
     def test_as_bytes_encoding(self, conn, encoding):
@@ -366,6 +405,8 @@ class TestLiteral:
     @pytest.mark.crdb_skip("composite")  # create type, actually
     @pytest.mark.parametrize("name", ["a-b", f"{eur}", "order", "foo bar"])
     def test_invalid_name(self, conn, name):
+        if conn.info.parameter_status("is_superuser") != "on":
+            pytest.skip("not a superuser")
         conn.execute(
             f"""
             set client_encoding to utf8;
@@ -457,6 +498,7 @@ class TestSQL:
 
     def test_as_string(self, conn):
         assert sql.SQL("foo").as_string(conn) == "foo"
+        assert sql.SQL("foo").as_string() == "foo"
 
     @pytest.mark.parametrize("encoding", ["utf8", crdb_encoding("latin9")])
     def test_as_bytes(self, conn, encoding):
@@ -465,6 +507,10 @@ class TestSQL:
 
         assert sql.SQL(eur).as_bytes(conn) == eur.encode(encoding)
 
+    def test_no_conn(self):
+        assert sql.SQL(eur).as_string() == eur
+        assert sql.SQL(eur).as_bytes() == eur.encode()
+
 
 class TestComposed:
     def test_class(self):
@@ -524,10 +570,12 @@ class TestComposed:
     def test_as_string(self, conn):
         obj = sql.Composed([sql.SQL("foo"), sql.SQL("bar")])
         assert obj.as_string(conn) == "foobar"
+        assert obj.as_string() == "foobar"
 
     def test_as_bytes(self, conn):
         obj = sql.Composed([sql.SQL("foo"), sql.SQL("bar")])
         assert obj.as_bytes(conn) == b"foobar"
+        assert obj.as_bytes() == b"foobar"
 
     @pytest.mark.parametrize("encoding", ["utf8", crdb_encoding("latin9")])
     def test_as_bytes_encoding(self, conn, encoding):
@@ -568,17 +616,21 @@ class TestPlaceholder:
     def test_as_string(self, conn, format):
         ph = sql.Placeholder(format=format)
         assert ph.as_string(conn) == f"%{format.value}"
+        assert ph.as_string() == f"%{format.value}"
 
         ph = sql.Placeholder(name="foo", format=format)
         assert ph.as_string(conn) == f"%(foo){format.value}"
+        assert ph.as_string() == f"%(foo){format.value}"
 
     @pytest.mark.parametrize("format", PyFormat)
     def test_as_bytes(self, conn, format):
         ph = sql.Placeholder(format=format)
-        assert ph.as_bytes(conn) == f"%{format.value}".encode("ascii")
+        assert ph.as_bytes(conn) == f"%{format.value}".encode()
+        assert ph.as_bytes() == f"%{format.value}".encode()
 
         ph = sql.Placeholder(name="foo", format=format)
-        assert ph.as_bytes(conn) == f"%(foo){format.value}".encode("ascii")
+        assert ph.as_bytes(conn) == f"%(foo){format.value}".encode()
+        assert ph.as_bytes() == f"%(foo){format.value}".encode()
 
 
 class TestValues:
@@ -593,7 +645,7 @@ class TestValues:
 
 def no_e(s):
     """Drop an eventual E from E'' quotes"""
-    if isinstance(s, memoryview):
+    if isinstance(s, (memoryview, bytearray)):
         s = bytes(s)
 
     if isinstance(s, str):
