@@ -3,14 +3,14 @@
 from __future__ import annotations
 
 from dataclasses import dataclass
-from typing import Any, Callable, Dict, Optional, Sequence, Tuple, Union
+from typing import Any, Callable, Sequence
 
 from psycopg import Connection, Cursor, ServerCursor, connect, rows
 from psycopg import AsyncConnection, AsyncCursor, AsyncServerCursor
 
 
 def int_row_factory(
-    cursor: Union[Cursor[Any], AsyncCursor[Any]]
+    cursor: Cursor[Any] | AsyncCursor[Any],
 ) -> Callable[[Sequence[int]], int]:
     return lambda values: values[0] if values else 42
 
@@ -22,7 +22,7 @@ class Person:
 
     @classmethod
     def row_factory(
-        cls, cursor: Union[Cursor[Any], AsyncCursor[Any]]
+        cls, cursor: Cursor[Any] | AsyncCursor[Any]
     ) -> Callable[[Sequence[str]], Person]:
         def mkrow(values: Sequence[str]) -> Person:
             name, address = values
@@ -45,12 +45,12 @@ def check_row_factory_cursor() -> None:
 
     cur1: Cursor[Any]
     cur1 = conn.cursor()
-    r1: Optional[Any]
+    r1: Any | None
     r1 = cur1.fetchone()
     r1 is not None
 
     cur2: Cursor[int]
-    r2: Optional[int]
+    r2: int | None
     with conn.cursor(row_factory=int_row_factory) as cur2:
         cur2.execute("select 1")
         r2 = cur2.fetchone()
@@ -70,12 +70,12 @@ async def async_check_row_factory_cursor() -> None:
 
     cur1: AsyncCursor[Any]
     cur1 = conn.cursor()
-    r1: Optional[Any]
+    r1: Any | None
     r1 = await cur1.fetchone()
     r1 is not None
 
     cur2: AsyncCursor[int]
-    r2: Optional[int]
+    r2: int | None
     async with conn.cursor(row_factory=int_row_factory) as cur2:
         await cur2.execute("select 1")
         r2 = await cur2.fetchone()
@@ -95,7 +95,7 @@ def check_row_factory_connection() -> None:
     """
     conn1: Connection[int]
     cur1: Cursor[int]
-    r1: Optional[int]
+    r1: int | None
     conn1 = connect(row_factory=int_row_factory)
     cur1 = conn1.execute("select 1")
     r1 = cur1.fetchone()
@@ -105,7 +105,7 @@ def check_row_factory_connection() -> None:
 
     conn2: Connection[Person]
     cur2: Cursor[Person]
-    r2: Optional[Person]
+    r2: Person | None
     conn2 = connect(row_factory=Person.row_factory)
     cur2 = conn2.execute("select * from persons")
     r2 = cur2.fetchone()
@@ -113,8 +113,8 @@ def check_row_factory_connection() -> None:
     with conn2.cursor() as cur2:
         cur2.execute("select 2")
 
-    cur3: Cursor[Tuple[Any, ...]]
-    r3: Optional[Tuple[Any, ...]]
+    cur3: Cursor[tuple[Any, ...]]
+    r3: tuple[Any, ...] | None
     conn3 = connect()
     cur3 = conn3.execute("select 3")
     with conn3.cursor() as cur3:
@@ -129,7 +129,7 @@ async def async_check_row_factory_connection() -> None:
     """
     conn1: AsyncConnection[int]
     cur1: AsyncCursor[int]
-    r1: Optional[int]
+    r1: int | None
     conn1 = await AsyncConnection.connect(row_factory=int_row_factory)
     cur1 = await conn1.execute("select 1")
     r1 = await cur1.fetchone()
@@ -139,7 +139,7 @@ async def async_check_row_factory_connection() -> None:
 
     conn2: AsyncConnection[Person]
     cur2: AsyncCursor[Person]
-    r2: Optional[Person]
+    r2: Person | None
     conn2 = await AsyncConnection.connect(row_factory=Person.row_factory)
     cur2 = await conn2.execute("select * from persons")
     r2 = await cur2.fetchone()
@@ -147,8 +147,8 @@ async def async_check_row_factory_connection() -> None:
     async with conn2.cursor() as cur2:
         await cur2.execute("select 2")
 
-    cur3: AsyncCursor[Tuple[Any, ...]]
-    r3: Optional[Tuple[Any, ...]]
+    cur3: AsyncCursor[tuple[Any, ...]]
+    r3: tuple[Any, ...] | None
     conn3 = await AsyncConnection.connect()
     cur3 = await conn3.execute("select 3")
     async with conn3.cursor() as cur3:
@@ -159,10 +159,10 @@ async def async_check_row_factory_connection() -> None:
 
 def check_row_factories() -> None:
     conn1 = connect(row_factory=rows.tuple_row)
-    v1: Tuple[Any, ...] = conn1.execute("").fetchall()[0]
+    v1: tuple[Any, ...] = conn1.execute("").fetchall()[0]
 
     conn2 = connect(row_factory=rows.dict_row)
-    v2: Dict[str, Any] = conn2.execute("").fetchall()[0]
+    v2: dict[str, Any] = conn2.execute("").fetchall()[0]
 
     conn3 = connect(row_factory=rows.class_row(Person))
     v3: Person = conn3.execute("").fetchall()[0]
